;       *************************
;       *    C L O C K   8 0    *
;       *    By Robert Doerr    *
;       *-----------------------*
;       *    Blinking  Clock    *
;       *************************
;
	ORG     6000H           ;Start address of program
;
;       Initilize program
;
START   DI                      ;Disable interrupts
	LD      SP,7FFFH        ;Set stack
	LD      A,(0125H)       ;Test for Model III
	CP      49H             ;Is Character an "I"?
	JR      Z,MODEL3        ;Jump if Model III
	LD      A,'a'           ;Get lower case "a"
	LD      (3C00H),A       ;Save on screen
	LD      A,(3C00H)       ;Get back from screen
	CP      'a'             ;Is it still an "a"?
	JR      Z,OKL           ;Jump if still "a"
	LD      HL,0            ;Op-Code for "NOP,NOP"
	LD      (PRSTR),HL      ;Zero out Jump
OKL     LD      A,49H           ;Create Model I Version
	LD      (WHAT+51),A     ;Modify startup Message
	LD      HL,0            ;Op-code for "NOP,NOP"
	LD      (ZAP1),HL       ;Zero out instruction
	LD      (ZAP2),HL       ;Zero out instruction
	LD      (ZAP3),HL       ;Zero out instruction
	LD      (ZAP4),HL       ;Zero out instruction
	LD      (ZAP5),HL       ;Zero out instruction
	LD      (ZAP6+2),HL     ;Zero out instruction
	LD      A,3AH           ;Op-Code for "LD A,(???)"
	LD      (XIT),A         ;Change Op-Code
	LD      (FDC),A         ;Change Op-Code
	LD      (ZAP6),A        ;Change Op-Code
	LD      A,37H           ;Msb of Interrupt address
	LD      (XIT+2),A       ;Change Byte
	LD      (FDC+2),A       ;Change Byte
	LD      (ZAP6+2),A      ;Change Byte
	LD      A,0E0H          ;Lsb of Interrupt address
	LD      (ZAP6+1),A      ;Set Lsb of address
	LD      A,40            ;Model I has 40 ticks
	LD      (TICKS+1),A     ;Change ticks
	LD      A,20            ;Half second
	LD      (HLFTIC+1),A    ;Change half second
;
;       Print startup Message
;       and get the time
;
MODEL3  CALL    CLS             ;Clear the screen
	LD      HL,WHAT         ;Message to print
	LD      DE,3C00H        ;1st line of screen
	CALL    PRSTR           ;Print Message
	LD      HL,TIME         ;Message to print
	LD      DE,3C80H        ;3rd line of screen
	CALL    PRSTR           ;Print Message
	CALL    INPUT           ;Input the Time
;
;       Do syntax check on input line
;
	LD      HL,TIMEBF       ;Start of Input buffer
	CALL    CHECKT          ;Check 2 digits
	LD      A,(HL)          ;Get next Character
	CP      ':'             ;Is it a colon?
	JP      NZ,ERROR        ;Error if no colon
	INC     HL              ;Bump pointer
	CALL    CHECKT          ;Check 2 digits
	LD      A,(HL)          ;Get next Character
	CP      ':'             ;Is it a colon?
	JP      NZ,ERROR        ;Error if no colon
	INC     HL              ;Bump pointer
	CALL    CHECKT          ;Check 2 digits
	LD      A,(HL)          ;Get next Character
	CP      '-'             ;Is it a dash?
	JR      Z,OKGET         ;Jump if it is
	CP      ' '             ;Is it a space?
	JR      Z,OKGET         ;Jump if it is
	CP      ':'             ;Is it a colon?
	JP      NZ,ERROR        ;Error if it isn't
OKGET   INC     HL              ;Bump pointer
	LD      A,(HL)          ;Get next Character
	CP      'A'             ;Is it an "A"?
	JR      Z,OKAM          ;Jump if it is
	CP      'a'             ;Is it an "a"?
	JR      NZ,CHEKP        ;If not check for "P"
OKAM    XOR     A               ;Set flag for "AM" time
	LD      (FLIP),A        ;Save the flag
	JR      CHEKM           ;Continue
CHEKP   CP      'P'             ;Is it a "P"?
	JR      Z,OKPM          ;Jump if it is
	CP      'p'             ;Is it a "p"?
	JP      NZ,ERROR        ;Error in time entry
OKPM    LD      A,1             ;Set flag for "PM" time
	LD      (FLIP),A        ;Save the flag
CHEKM   INC     HL              ;Bump pointer
	LD      A,(HL)          ;Get next Character
	CP      'M'             ;Is it an "M"?
	JR      Z,OKTIM         ;Jump if it is
	CP      'm'             ;Is it an "m"?
	JP      NZ,ERROR        ;Error in time entry
;
;       Convert the ASCII time to Binary
;
OKTIM   LD      HL,TIMEBF       ;Start of input buffer
	CALL    EVAL            ;Evaluate number
	CP      13              ;Check for 0-12
	JP      NC,ERROR        ;Value too big. Error
	OR      A               ;Check for zero
	JP      Z,ERROR ;Value too small. Error
	LD      (HOURS),A       ;Save hours
	INC     HL              ;Bump pointer
	CALL    EVAL            ;Evaluate number
	CP      60              ;Check for 0-59
	JP      NC,ERROR        ;Value too big. Error
	LD      (MINITE),A      ;Save minutes
	INC     HL              ;Bump pointer
	CALL    EVAL            ;Evaluate number
	CP      60              ;Check for 0-59
	JP      NC,ERROR        ;Value too big. Error
	LD      (SECOND),A      ;Save seconds
	XOR     A               ;Load A with zero
	LD      (CLK),A         ;Zero out ticks
;
;       Set up interrupt routine
;
	IM      1               ;Set interrupt mode
ZAP1    OUT     (0E0H),A        ;Reset Model III
ZAP2    LD      A,38H           ;Value to send
ZAP3    OUT     (0ECH),A        ;Send value
	LD      A,0C3H          ;Code for "JP ???"
	LD      (4012H),A       ;Route interrupt Vector
	LD      HL,ISR          ;Address for Vector
	LD      (4013H),HL      ;Save address
ZAP4    LD      A,4             ;Value to send
ZAP5    OUT     (0E0H),A        ;Send value
	EI                      ;Enable interrupts
;
;       Main program loop
;
LOOP    LD      A,(3840H)       ;Keyboard row with BREAK
	BIT     2,A             ;Check BREAK key bit
	JR      Z,CONT          ;Jump if not pressed
RELES   LD      A,(3840H)       ;Keyboard row with BREAK
	BIT     2,A             ;Check BREAK key bit
	JR      NZ,RELES        ;Jump if holding key
	JP      MODEL3          ;Restart program
CONT    LD      A,(CLK)         ;Get ticks
HLFTIC  CP      15              ;Half second
	CALL    Z,BLKOUT        ;Blank colons if half
	OR      A               ;Check for zero
	JR      NZ,LOOP         ;Jump if no change
;
;       Convert binary time to decimal
;
CLOCK   LD      IY,TIMEBF       ;Input buffer
	LD      A,(HOURS)       ;Get the hour
	CALL    GETTWO          ;Convert to ASCII
	LD      A,(MINITE)      ;Get the minute
	CALL    GETTWO          ;Convert to ASCII
	LD      A,(SECOND)      ;Get the second
	CALL    GETTWO          ;Convert to ASCII
;
;       Clear screen buffer
;
	LD      HL,BUFFER       ;Start of buffer
	LD      DE,BUFFER+1     ;2nd character in buffer
	LD      BC,3FFH         ;Number of bytes to move
	LD      (HL),80H        ;Save blank in buffer
	LDIR                    ;Clear the buffer
;
;       Draw the Clock in the screen buffer
;
	LD      IY,TIMEBF       ;Point to time
	LD      HL,BUFFER+200   ;Location of 1st digit
	LD      DE,6            ;Offset for characters
	LD      B,8             ;Number of characters
GETVAL  LD      A,(IY)          ;Get digit
	SUB     2FH             ;Get number 1-11
	PUSH    HL              ;Save HL register
	PUSH    DE              ;Save DE register
	PUSH    BC              ;Save BC register
	CALL    DISDIG          ;Display a digit
	POP     BC              ;Restore BC register
	POP     DE              ;Restore DE register
	POP     HL              ;Restore HL register
	ADD     HL,DE           ;Add offset
	INC     IY              ;Point to next character
	DJNZ    GETVAL          ;Continue until done
;
;       Print AM or PM in screen buffer
;
	LD      DE,BUFFER+195   ;Location for AM and PM
	LD      HL,AM           ;AM message
	LD      A,(FLIP)        ;Get flag
	OR      A               ;AM or PM?
	JR      Z,PRINT         ;Print it if AM time
	INC     HL              ;Point to PM message
	INC     HL              ;Bump pointer
	INC     HL              ;Bump pointer
PRINT   CALL    PRSTR           ;Print it
;
;       Move screen buffer to screen
;
	LD      HL,BUFFER       ;Screen buffer
	LD      DE,3C00H        ;Screen
	LD      BC,3FFH         ;Buffer length
	LDIR                    ;Move it!
	JP      LOOP            ;Continue with program
;
;       Clear screen routine
;
CLS     LD      HL,3C00H        ;Start of screen
	LD      DE,3C01H        ;2nd spot on screen
	LD      BC,3FFH         ;Length of screen minus 1
	LD      (HL),20H        ;Put blank in first spot
	LDIR                    ;Clear screen
	RET                     ;And leave.
;
;       Print digit in buffer
;
DISDIG  PUSH    HL              ;Save HL register
	LD      HL,DATAB-7      ;Start of number data
	LD      DE,7            ;Offset for number data
	LD      B,A             ;Get number
CALCST  ADD     HL,DE           ;Add offset
	DJNZ    CALCST          ;Continue until done
	PUSH    HL              ;Save HL register
	POP     IX              ;Restore into IX register
	POP     HL              ;Restore HL register
	LD      DE,59           ;Line offset
	LD      B,7             ;Number of Lines
DISLOP  PUSH    BC              ;Save BC register
	LD      A,(IX)          ;Get row data
	RLCA                    ;Shift it
	RLCA                    ;Shift it again
	RLCA                    ;And again
	LD      B,5             ;Number of bits to print
DISROW  RLCA                    ;Get bit in carry flag
	JR      NC,CONTWT       ;Skip if nothing there
	LD      (HL),191        ;Turn on block in buffer
CONTWT  INC     HL              ;Update pointer
	DJNZ    DISROW          ;Finish line
	ADD     HL,DE           ;Get next line
	INC     IX              ;Point to new data
	POP     BC              ;Restore BC register
	DJNZ    DISLOP          ;Finish character
	RET                     ;Return
;
;       Routine to erase both colons
;
BLKOUT  LD      HL,15572        ;Address of first colon
	CALL    BLANK           ;Erase first colon
	LD      HL,15590        ;Address of second colon
BLANK   LD      B,7             ;Number of lines to erase
BLKLP   PUSH    BC              ;Save line count
	LD      B,5             ;Number of bytes on line
BLKLP2  LD      (HL),128        ;Erase one character
	INC     HL              ;Point to next character
	DJNZ    BLKLP2          ;Finish line
	POP     BC              ;Restore line count
	LD      DE,59           ;Offset for next line
	ADD     HL,DE           ;Add offset
	DJNZ    BLKLP           ;Finish character
	RET                     ;And return
;
;       Convert the binary number in A register
;       into a two digit ASCII number and save
;
GETTWO  LD      B,0FFH          ;Load B with -1
LOOPTW  INC     B               ;Add one to tens counter
	SUB     10              ;Subtract ten
	JP      P,LOOPTW        ;Jump if not below zero
	ADD     A,3AH           ;Least significant digit
	LD      (IY+1),A        ;Save it in buffer
	LD      A,30H           ;offset for ASCII digit
	ADD     A,B             ;Most significant digit
	LD      (IY),A          ;Save it in buffer
	INC     IY              ;Bump pointer
	INC     IY              ;Bump pointer
	INC     IY              ;Bump pointer
	RET                     ;And return
;
;       Evaluate 2-Digit number at (HL)
;
EVAL    CALL    GETDID          ;Get binary digit
	ADD     A,A             ;Digit times 2
	LD      C,A             ;Save in C register
	ADD     A,A             ;Digit times 4
	ADD     A,A             ;Digit times 8
	ADD     A,C             ;Now we have digit times 10
	LD      C,A             ;Save in C register
	CALL    GETDID          ;Get binary digit
	ADD     A,C             ;Add to rest for number
	RET                     ;Return result in A register
GETDID  LD      A,(HL)          ;Get ASCII digit
	SUB     30H             ;Convert to binary digit
	INC     HL              ;Bump pointer
	RET                     ;Return binary digit
;
;       Check two digits
;
CHECKT  CALL    CHECKO          ;Check digit
	INC     HL              ;Bump pointer
	CALL    CHECKO          ;Check digit
	INC     HL              ;Bump pointer
	RET                     ;Return
;
;       Check for legal digit
;
CHECKO  LD      A,(HL)          ;Get character
	CP      '0'             ;Check for zero
	JR      C,ERROR2        ;Jump if error
	CP      ':'             ;Check for less than ":"
	RET     C               ;All ok!
;
;       Illegal time input
;
ERROR2  POP     HL              ;Restore stack
	POP     HL              ;Restore stack
ERROR   LD      HL,NOGOOD       ;Error message
	LD      DE,3D00H        ;5th line down
	CALL    PRSTR           ;Print it
SCANEN  CALL    0049H           ;Wait for character
	CP      0DH             ;Is it <ENTER>?
	JR      NZ,SCANEN       ;If not try again
	JP      MODEL3          ;Get new input
;
;       Print message routine
;
PRSTR   JR      SAME            ;Jump to second routine
PRLOP   LD      A,(HL)          ;Get character
	OR      A               ;Check for end
	RET     Z               ;Return if done
	CP      'a'             ;Check if less than "a"
	JR      C,OKPRT         ;Jump if less
	CP      7BH             ;Check if less than 7BH
	JR      NC,OKPRT        ;Jump if greater
	SUB     20H             ;Convert to UPPER CASE
OKPRT   LD      (DE),A          ;Save character
	INC     DE              ;Update Destination
	INC     HL              ;Update source
	JR      PRLOP           ;Continue printing
SAME    LD      A,(HL)          ;Get character
	OR      A               ;Check for end
	RET     Z               ;Return if done
	LD      (DE),A          ;Save character
	INC     DE              ;Update destination
	INC     HL              ;Update source
	JR      SAME            ;Continue printing
;
;       Line input routine
;
INPUT   LD      B,11            ;Maximum input
	LD      HL,3C80H+32     ;Screen location
	LD      DE,TIMEBF       ;Time buffer
	PUSH    BC              ;Save BC register
	PUSH    HL              ;Save HL register
INLOP1  LD      (HL),2EH        ;Print period on screen
	INC     HL              ;Point to next place
	DJNZ    INLOP1          ;Continue until done
	POP     HL              ;Restore HL register
	POP     BC              ;Restore BC register
	PUSH    DE              ;Save DE register
	LD      A,20H           ;Blank character
INLOP2  LD      (DE),A          ;Clear buffer
	INC     DE              ;Point to next place
	DJNZ    INLOP2          ;Continue until done
	POP     DE              ;Restore DE register
STATIN  LD      (HL),143        ;Print cursor
GETIHR  PUSH    DE              ;Save DE register
	CALL    0049H           ;Wait for key press
	POP     DE              ;Restore DE register
	CP      13              ;Check for <ENTER>
	JR      Z,ENTERC        ;Jump if <ENTER>
	CP      24              ;Check for <SHIFT> <--
	JR      Z,INPLS         ;Jump if it is
	CP      8               ;Check for <--
	JR      Z,BACKSP        ;Jump if it is
	CP      32              ;Check if less than 32
	JR      C,GETIHR        ;Jump if Illegal
	CP      123             ;Check if more than 122
	JR      NC,GETIHR       ;Jump if Illegal
	LD      C,A             ;Save A register in C
	LD      A,B             ;Get number of characters
	CP      11              ;Check if full
	JR      Z,GETIHR        ;Jump if full
	INC     B               ;Got one more
	LD      A,C             ;Get A back from C
	LD      (DE),A          ;Save in buffer
	INC     DE              ;Point to next space
	LD      A,(PRSTR)       ;Get Upper/Lower flag
	OR      A               ;Test for zero
	LD      A,C             ;Get A back from C
	JR      NZ,OPRNT        ;Jump if Upper/Lower case
	CP      'a'             ;Check if lower case
	JR      C,OPRNT         ;Jump if not
	SUB     20H             ;Convert to UPPER CASE
OPRNT   LD      (HL),A          ;Print on screen
	INC     HL              ;Point to next space
	JR      STATIN          ;Get next character
ENTERC  LD      A,B             ;Get number of characters
	CP      11              ;See if maximum
	JR      NZ,PEREXT       ;Jump if no maximum
	LD      (HL),20H        ;Blank cursor
	RET                     ;And return
PEREXT  LD      (HL),2EH        ;Blank cursor
	LD      A,20H           ;Get blank space
	LD      (DE),A          ;Put in buffer
	RET                     ;Return
BACKSP  LD      A,B             ;Get number of characters
	OR      A               ;Test for zero
	JR      Z,GETIHR        ;Jump if nothing there
	CP      11              ;Test for maximum
	JR      NZ,PERD         ;Jump if not maximum
	LD      (HL),20H        ;Blank out cursor
	DEC     HL              ;Backup pointer
	JR      BACKAL          ;Continue
PERD    LD      (HL),2EH        ;Print period
	DEC     HL              ;Backup pointer
	LD      A,20H           ;Blank character
	LD      (DE),A          ;Put blank in buffer
BACKAL  DEC     DE              ;Backup pointer
	DEC     B               ;Backup character count
	JR      STATIN          ;Continue
INPLS   LD      A,B             ;Get number of characters
	OR      A               ;Test for zero
	JR      Z,STATIN        ;Jump if zero
	LD      (HL),20H        ;Blank cursor
	JP      INPUT           ;Start over
;
;       Messages and number data
;
WHAT    DEFM    "CLOCK 80 -- 01/06/83 1983 By Robert Doerr -- Model 3 Version 1.2"
	DEFB    0       ;Terminator
TIME    DEFM    "Enter time (HH:MM:SS-AM/PM) -->"
	DEFB    0       ;Terminator
NOGOOD  DEFM    "Illegal time specification.  Press <ENTER>."
	DEFB    0       ;Terminator
AM      DEFM    "AM"
	DEFB    0
	DEFM    "PM"
	DEFB    0       ;Terminator
TIMEBF  DEFS    12      ;Input buffer
FLIP    DEFB    0       ;AM/PM flag
CLK     DEFB    0       ;Ticks
SECOND  DEFB    0       ;Seconds
MINITE  DEFB    0       ;Minutes
HOURS   DEFB    0       ;Hours
DATAB   DEFB    14      ;data for zero
	DEFB    17
	DEFB    19
	DEFB    21
	DEFB    25
	DEFB    17
	DEFB    14
ONE     DEFB    4       ;data for one
	DEFB    12
	DEFB    4
	DEFB    4
	DEFB    4
	DEFB    4
	DEFB    14
TWO     DEFB    14      ;data for two
	DEFB    17
	DEFB    1
	DEFB    14
	DEFB    16
	DEFB    16
	DEFB    31
THREE   DEFB    14      ;data for three
	DEFB    17
	DEFB    1
	DEFB    6
	DEFB    1
	DEFB    17
	DEFB    14
FOUR    DEFB    2       ;data for four
	DEFB    6
	DEFB    10
	DEFB    18
	DEFB    31
	DEFB    2
	DEFB    2
FIVE    DEFB    31      ;data for five
	DEFB    16
	DEFB    30
	DEFB    1
	DEFB    1
	DEFB    17
	DEFB    14
SIX     DEFB    6       ;data for six
	DEFB    8
	DEFB    16
	DEFB    30
	DEFB    17
	DEFB    17
	DEFB    14
SEVEN   DEFB    31      ;data for seven
	DEFB    1
	DEFB    2
	DEFB    4
	DEFB    8
	DEFB    16
	DEFB    16
EIGHT   DEFB    14      ;data for eight
	DEFB    17
	DEFB    17
	DEFB    14
	DEFB    17
	DEFB    17
	DEFB    14
NINE    DEFB    14      ;data for nine
	DEFB    17
	DEFB    17
	DEFB    15
	DEFB    1
	DEFB    2
	DEFB    12
COLON   DEFB    0       ;data for colon
	DEFB    14
	DEFB    14
	DEFB    0
	DEFB    14
	DEFB    14
	DEFB    0
BUFFER  DEFS    400H    ;Buffer for screen
;
;       Interrupt routine
;
ISR     PUSH    HL              ;Save HL register
	PUSH    AF              ;Save AF register
ZAP6    IN      A,(0ECH)        ;Clear interrupt
	IN      A,(0E0H)        ;Read interrupt status
	BIT     6,A             ;Test request flag
	JR      NZ,FDC          ;Jump if FDC request
	BIT     7,A             ;Test interrupt flag
	JR      Z,XIT           ;Exit if invalid
INTER   LD      HL,CLK          ;Point to ticks
	INC     (HL)            ;Increment ticks
	LD      A,(HL)          ;Get ticks
TICKS   CP      30              ;30 Ticks (Model III)
	JR      C,XIT           ;Not there yet
	XOR     A               ;Put Zero in A register
	LD      (HL),A          ;Reset ticks
	INC     HL              ;Point to seconds
	INC     (HL)            ;Increment seconds
	LD      A,(HL)          ;Get seconds
	CP      60              ;Increment minutes?
	JR      C,XIT           ;Nope.
	XOR     A               ;Put zero in A register
	LD      (HL),A          ;Reset seconds
	INC     HL              ;Point to minutes
	INC     (HL)            ;Increment minutes
	LD      A,(HL)          ;Get minutes
	CP      60              ;Increment hours?
	JR      C,XIT           ;Nope.
	XOR     A               ;Put zero in A register
	LD      (HL),A          ;Reset minutes
	INC     HL              ;Point to hours
	INC     (HL)            ;Increment hours
	LD      A,(HL)          ;Get hours
	CP      12              ;Check for swap
	JR      Z,SWAP          ;Swap AM and PM if 12
	CP      13              ;Flip counter?
	JR      C,XIT           ;Nope.
	LD      A,1             ;Set for 1 hour
	LD      (HL),A          ;Reset hours
	JR      XIT             ;Leave
SWAP    LD      A,(FLIP)        ;Get AM and PM flag
	XOR     1               ;Reverse it
	LD      (FLIP),A        ;Save new flag
XIT     IN      A,(0E0H)        ;Reset latch
	NOP                     ;No operation
	POP     AF              ;Restore AF register
	POP     HL              ;Restore HL register
	EI                      ;Enable interrupts
	RET                     ;Leave
FDC     IN      A,(0ECH)        ;Reset FDC request
	NOP                     ;No operation
	JR      INTER           ;Continue
