'******************************************************************************
'*  io.spin - handle input, output and break requests sent by yaz80
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: io.spin,v 1.28 2010-03-31 05:33:17 pm Exp $
'******************************************************************************
CON

' Define this to check for contiguous image files
#define	CHECK_CONTIG

	io_cmd_out		=	$01		' I/O emulator command output
	io_cmd_in		=	$02		' I/O emulator command input
	io_cmd_break		=	$03		' I/O emulator command break
	io_cmd_kbd		=	$04		' I/O emulator command read (TRS-80 keyboard)

	constat			=	$10		' (16)sio port 1 status port
	condata			=	$11		' (17)sio port 1 data port
	punstat			=	$12		' (18)sio port 2 status port
	pundata			=	$13		' (19)sio port 2 data port

	myname_port		=	$70		' hopefully not used by anything else
	port2_Off		=	$71		' a little CP/M program sends an OUT to port &H71
	Forecolor		=	$72		' set foreground color
	Backcolor		=	$73		' set background color
	keystat			=	$74		' test just keyboard status (constat does this, too)
	keydata			=	$75		' get just the keyboard character

	port2_io		=	$78		' for input and output to port 2
	port2_test		=	$7D		' to test if a byte is ready
	port2_baud		=	$7C		' set baud rate multiple of 1200 eg 1=1200, 16=19200, 96=115200

	hdsk_reset		=	$1		' command to reset controller
	hdsk_read		=	$2		' read command
	hdsk_write		=	$3		' write command
	hdsk_param		=	$4		' get hard disk parameters
	hdsk_port		=	$fd		' control port for simulated hard disk

	simh_setbankselect	=	12		' SIMH command to set memory bank
	simh_reset		=	14		' SIMH reset command
	simh_hasbankedmemory	=	18		' SIMH command to check for banked memory
	simh_stoptimer		=	22		' SIMH stop timer interrupt
	simh_port		=	$fe		' control port for simulated hard disk

	baud			=	115200		' baud rate

	_nul			=	0		' clear screen code
	_cr			=	13		' carriage return
	dir_none		=	0		' no i/o direction selected
	dir_out			=	1		' expecting client to write to the port
	dir_in			=	2		' expecting client to read from the port

VAR
	long	cog
	long	io_command
	long	iobuff[2*512/4]				' I/O buffer (need 2 because of FAT12)
	long	screen[80*25/4]
#ifdef HAVE_VGA
	long	vsync
	long	colors[64]
	byte	cursor[6]
#endif HAVE_VGA
	byte	name[12+1]
	byte	datetime[20+1]

OBJ
	fdx	: "pcFullDuplexSerial4FC"
	fat	: "fatfs"
#ifdef TriBladeProp
	xmm	: "TriBladeProp_Blade2_204"		' Extended memory manager
#endif TriBladeProp
#ifdef RamBladeProp
	xmm	: "RamBlade_203_002"			' Extended memory manager
#endif RamBladeProp
#ifdef DracBladeProp
	xmm	: "Dracblade"				' Extended memory manager
#endif DracBladeProp
#ifdef HAVE_KBD
	kbd	: "Keyboard"
#endif HAVE_KBD
#ifdef HAVE_VGA
	vga	: "VGA_HiRes_Text"
#endif HAVE_VGA

PUB start(SI, SO, SI2, SO2, doPin, clkPin, diPin, csPin, kbDpin, kbCpin, vgaPin) : okay | i, dent, sec, size
	stop

	clk_freq := clkfreq
	waitcnt(CNT + clkfreq * 2)

	iobuff_ptr := @iobuff

	okay := \fat.start(iobuff_ptr, doPin, clkPin, diPin, csPin)
	if okay < 0
		return -1
	sd_command := fat.get_SPI_command
	sd_blocknr := sd_command + 4
	sd_address := sd_command + 8

	fdx.AddPort(0, SI, SO, -1, -1, 0, 0, baud)
	if SI2 <> -1
		fdx.AddPort(1, SI2, SO2, -1, -1, 0, 0, baud)

	okay := fdx.Start
	if okay < 0
		return
	tx_head := fdx.get_tx_head
	tx_tail := fdx.get_tx_tail
	tx_buffer := fdx.get_tx_buffer
	rx_head := fdx.get_rx_head
	rx_tail := fdx.get_rx_tail
	rx_buffer := fdx.get_rx_buffer

	if SI2 <> -1
		tx_head1 := fdx.get_tx_head1
		tx_tail1 := fdx.get_tx_tail1
		tx_buffer1 := fdx.get_tx_buffer1
		rx_head1 := fdx.get_rx_head1
		rx_tail1 := fdx.get_rx_tail1
		rx_buffer1 := fdx.get_rx_buffer1

#ifdef HAVE_TV
	if tvPin == -1
		return -1
	tv.start(tvPin, @screen)
	tv_command := tv.get_command
#endif
	str(string(27,"[2J",27,"[H","YAZ80 - Yet Another Z80...",13,10))

#ifdef HAVE_KBD
	if kbDpin == -1 or kbCpin == -1
		return -1
	okay := \kbd.startx(kbDpin, kbCpin, %100, 40)
	if okay < 0
		return -1
	kbd_tail := kbd.tail_ptr
	kbd_head := kbd.head_ptr
	kbd_keys := kbd.keys_ptr
#endif

#ifdef HAVE_VGA
	if vgaPin == -1
		return -1
	okay := \vga.start(vgaPin, @screen, @colors, @cursor, @vsync)
	if okay > 0
		return -1
	longfill(@colors, %%0020_3300, 64)
	cursor[0] := 0		' cursor x
	cursor[1] := 0		' cursor y
	cursor[2] := %110	' on, blink slow
#endif
	okay := xmm.start
	if okay < 0
		return -1
	xmm_command := xmm.rendezvous
	xmm_hubaddr := xmm_command + 1*4
	xmm_ramaddr := xmm_command + 2*4
	xmm_blocklen := xmm_command + 3*4
	xmm_errx := xmm_command + 4*4

	okay := fat.mount
	if okay < 0
		failure(string("FAT mount"),okay)
		return
	str(string("Volume serial #"))
	hex(fat.get_serialno >> 16, 4)
	chr("-")
	hex(fat.get_serialno, 4)
	str(string(", label '"))
	str(fat.get_volume_label)
	str(string("'",13,10))

	read_boot_dsk(string("BOOT.DSK"))

	map_disk(@lba_drive_a, string("A.DSK"))
	map_disk(@lba_drive_b, string("B.DSK"))
	map_disk(@lba_drive_c, string("C.DSK"))
	map_disk(@lba_drive_d, string("D.DSK"))
	map_disk(@lba_drive_e, string("E.DSK"))
	map_disk(@lba_drive_f, string("F.DSK"))
	map_disk(@lba_drive_g, string("G.DSK"))
	map_disk(@lba_drive_h, string("H.DSK"))

	okay := cog := COGNEW(@entry, @io_command) + 1
	if okay =< 0
		failure(string("I/O"),okay)
		return -1
	success(string("I/O"))
	return 0

PUB stop
'' Stop CPU simulator - frees a cog
	if cog
		cogstop(cog~ - 1)


PUB chr(ch)
	fdx.tx(0,ch)

PUB str(str_ptr) | i
	repeat i from 0 to strsize(str_ptr)-1
		chr(byte[str_ptr][i])

PUB crlf
	chr(13)
	chr(10)

PUB dec_size(value)
	if value > 1048576
		dec(value/1048576)
		chr(".")
		dec(10*(value // 1048576) / 1048576)
		str(string("MB"))
	elseif value > 1024
		dec(value/1024)
		chr(".")
		dec(10*(value // 1024) / 1024)
		str(string("KB"))
	else
		dec(value)

PUB dec(value) | _i
'' Print a decimal number

	if value < 0
		-value
		chr("-")

	_i := 1_000_000_000

	repeat 10
		if value => _i
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			chr("0")
		_i /= 10

PUB hex(value, digits)
'' Print a hexadecimal number

	value <<= (8 - digits) << 2
	repeat digits
		chr(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)
'' Print a binary number

	value <<= 32 - digits
	repeat digits
		chr((value <-= 1) & 1 + "0")

PRI failure(who, val)
	str(who)
	str(string(" failed: "))
	dec(val)
	crlf

PRI success(who) | i, n, n1, n2, n3, n4, n5, n6, n7
	str(who)
	str(string(" initialized, "))
	dec(free_cogs)
	str(string(" cogs free.", 13, 10))

PUB free_cogs : free | i, cogs[8], jmp_0
	jmp_0 := %010111_0001_1111_000000000_000000000
	repeat while (cogs[free] := cognew(@jmp_0, 0)) => 0 
		free++
	if free
		repeat i from 0 to free - 1
			cogstop(cogs[i])
	return free

PRI read_boot_dsk(fname_ptr) | dent, size, sec
	dent := fat.find_file(fname_ptr)
	if dent == 0
		failure(fname_ptr,dent)
		return -1
	success(fname_ptr)
	size := fat.get_dent_size(dent)
	sec := fat.get_dent_first_sector(dent)
	fat.get_dent_name(dent, @name)
	fat.get_dent_date_time(dent, @datetime)
	str(@name)
	str(string(": sector "))
	dec(sec)
	str(string(9,"size "))
	dec_size(size)
	chr(9)
	str(@datetime)
	crlf

	fat.read_sector(sec)
	xmm.DoCmd("W", iobuff_ptr, $FF00, 256)
	long[iobuff_ptr] := $00FF00C3
	xmm.DoCmd("W", iobuff_ptr, $0000, 4)


PRI map_disk(lbaptr, fname_ptr) : okay | dent, size, clus, exp, err, sec
	dent := fat.find_file(fname_ptr)
	if dent < 0
		failure(fname_ptr, dent)
		return -1
	sec := fat.get_dent_first_sector(dent)
	size := fat.get_dent_size(dent)
	long[lbaptr] := sec
	fat.get_dent_name(dent, @name)
	fat.get_dent_date_time(dent, @datetime)
	str(@name)
	str(string(9,"sector "))
	dec(sec)
	str(string(9,"size "))
	dec_size(size)
	chr(9)
	str(@datetime)
#ifdef CHECK_CONTIG
	clus := fat.get_dent_cluster(dent)
	exp := clus + 1
	err := 0
	repeat while clus := fat.fat_next(clus)
		if clus <> exp
			err++
			quit
		exp := exp + 1
	if err
		str(string(" not contiguous!"))
	else
		str(string(" contiguous - okay."))
#endif
	crlf

PUB command_addr
'' Return the address of the 3 longs block
	return @io_command

DAT
'---------------------------------------------------------------------------------------------------------

		org	0

'---------------------------------------------------------------------------------------------------------
entry
iocmd_ptr	mov	iocmd_ptr, PAR
iocmd		mov	lba_iobuff, minus1
		jmp	#run

' pcFullDuplexSerial4FC
tx_head		long	0-0
tx_tail		long	0-0
tx_buffer	long	0-0
rx_head		long	0-0
rx_tail		long	0-0
rx_buffer	long	0-0

tx_head1	long	0-0
tx_tail1	long	0-0
tx_buffer1	long	0-0
rx_head1	long	0-0
rx_tail1	long	0-0
rx_buffer1	long	0-0

fdx_head	long	0
fdx_tail	long	0
fdx_temp	long	0

#ifdef HAVE_TV
tv_command	long	0-0
tv_temp		long	0
#endif

#ifdef HAVE_KBD
kbd_tail	long	0-0
kbd_head	long	0-0
kbd_keys	long	0-0
kbd_ctrl	long	512
#endif HAVE_KBD

xmm_command	long	0-0
xmm_hubaddr	long	0-0
xmm_ramaddr	long	0-0
xmm_blocklen	long	0-0
xmm_errx	long	0-0

sd_command	long	0-0
sd_blocknr	long	0-0
sd_address	long	0-0
iobuff_ptr	long	0-0

minus1		long	$ffffffff
low_24bits	long	$00ffffff
low_16bits	long	$0000ffff
msg_break_ptr	long	@@@msg_break

addr		long	0
port		long	0
data		long	0
null		long	0

simh_command	long	0		' currently active command
simh_status	long	0		' status byte
simh_params	long	@@@simh_params_	' pointer to the simh io buffer
simh_params_ptr	long	0		' pointer into simh_params
simh_params_cnt	long	0		' count of parameters expected

hdsk_command	long	0
hdsk_status	long	0
hdsk_lba_drives	long	@@@lba_drive_a
hdsk_drive	long	@@@hdsk_drive_	' parameters of read/write command
hdsk_sector	long	@@@hdsk_sector_
hdsk_track_lo	long	@@@hdsk_track_lo_
hdsk_track_hi	long	@@@hdsk_track_hi_
hdsk_dma_lo	long	@@@hdsk_dma_lo_
hdsk_dma_hi	long	@@@hdsk_dma_hi_

hdsk_dpb	long	@@@hdsk_dpb_

hdsk_dir	long	0
hdsk_params_ptr	long	0		' pointer into hdsk_params in hub RAM
hdsk_params_cnt	long	0		' count of parameters expected

boot0		long	0-0		' logical block address of boot sector
lba0		long	0-0		' logical block address of CP/M
lba		long	0		' logical block translated from track*32+sector
lba_iobuff	long	0		' logical block that is currently in iobuff
sector_offs	long	0		' offset of current sector in iobuff_ptr
clk_freq	long	0-0		' set by Spin code
iobuff_dirty	long	0
time_dirty	long	0

t1		long	0
t2		long	0
t3		long	0

strptr		long	0
hexval		long	0
hexcnt		long	0

run

command
		cmp	iobuff_dirty, #0	WZ		' is the iobuffer written to?
	if_nz	cmp	time_dirty, cnt		WC		' timeout reachead?
  if_nz_and_c	call	#sd_write_lba				' yes, write the buffer and clear the dirty flag
		rdlong	iocmd, iocmd_ptr
		mov	t1, iocmd
		shr	t1, #24			WZ		' inspect bits 31..24: zero?
	if_z	jmp	#command				' yes, loop waiting for a command.

		cmp	t1, #io_cmd_out		WZ
	if_z	jmp	#output
		cmp	t1, #io_cmd_in		WZ
	if_z	jmp	#input
		cmp	t1, #io_cmd_break	WZ
	if_z	jmp	#break
command_0
		and	iocmd, low_24bits			' clear command byte in bits 31..24
		wrlong	iocmd, iocmd_ptr
		jmp	#command

'*******************************************
'
' Read i/o port and data to write to
'
output
		mov	port, iocmd				' extract port from bits 23..8
		shr	port, #8				' to bits 15..0
		and	port, low_16bits			' mask bits 15..0
		and	port, #$ff				' mask lower 8 bits 7..0 of port address
		mov	data, iocmd				' extract data
		and	data, #$ff				' mask 8 data bits
		cmp	port, #simh_port 	WZ		' SIMH port?
	if_z	jmp	#ouput_simh
		cmp	port, #hdsk_port WZ			' hard disk port?
	if_z	jmp	#ouput_hdsk
		cmp	port, #condata WZ			' console output?
	if_z	jmp	#output_conout
		cmp	port, #punstat WZ			' punch status output - really?
	if_z	jmp	#output_punstat
		cmp	port, #pundata WZ			' punch output?
	if_z	jmp	#output_pundata
		jmp	#command_0

'*******************************************
'
' Write data to the console,
' writes to the FullDuplexSerial tx_buffer
'
output_conout
#ifdef HAVE_TV
:wait		rdlong	tv_temp, tv_command	WZ
	if_nz	jmp	#:wait
		wrlong	data, tv_command
#endif HAVE_TV
		call	#tx
		jmp	#command_0

'*******************************************
'
' Write to the FullDuplexSerial tx_buffer
'
tx
:wait
		rdlong	fdx_head, tx_head
		mov	fdx_temp, fdx_head
		add	fdx_head, #1
		rdlong	fdx_tail, tx_tail
		and	fdx_head, #15
		cmp	fdx_head, fdx_tail	WZ
	if_z	jmp	#:wait					' wait until there's room in the buffer
		mov	addr, tx_buffer
		add	addr, fdx_temp
		wrbyte	data, addr
		wrlong	fdx_head, tx_head			' increment the tx_head
tx_ret
		ret

'*******************************************
'
' Output a byte in hexval to the console
'
conout_hex_b
		mov	hexcnt, #2
		shl	hexval, #24
		jmp	#conout_hex

'*******************************************
'
' Output a word in hexval to the console
'
conout_hex_w
		mov	hexcnt, #4
		shl	hexval, #16
		jmp	#conout_hex

'*******************************************
'
' Output a long in hexval to the console
'
conout_hex_l
		mov	hexcnt, #8
conout_hex
		mov	data, hexval
		shr	data, #28
		cmp	data, #10	WZ, WC
	if_b	add	data, #48
	if_ae	add	data, #55
		call	#tx
		shl	hexval, #4
		djnz	hexcnt, #conout_hex
conout_hex_ret
conout_hex_l_ret
conout_hex_w_ret
conout_hex_b_ret
		ret

'*******************************************
'
' Output a space character to the console
'
conout_space
		mov	data, #32
		call	#tx
conout_space_ret
		ret

'*******************************************
'
' Output a carriage return to the console
'
conout_crlf
		mov	data, #13
		call	#tx
		mov	data, #10
		call	#tx
conout_crlf_ret
		ret

'*******************************************
'
' Output a string at strptr to the console
'
conout_string
		rdbyte	data, strptr	WZ
		add	strptr, #1
	if_z	jmp	conout_string_ret
		call	#tx
		jmp	#conout_string
conout_string_ret
		ret

'*******************************************
'
' Output to the punch status - needed?
'
output_punstat
		mov	data, #0
		jmp	#command_0

'*******************************************
'
' Output to the punch. TODO: serial port 2?
'
output_pundata
		jmp	#command_0

'*******************************************
'
' Handle SIMH compatible command
'
ouput_simh
		tjnz	simh_params_ptr, #output_simh_params

		cmp	data, #simh_reset WZ	' reset what?
	if_z	jmp	#output_simh_reset
		cmp	data, #simh_hasbankedmemory WZ
	if_z	jmp	#output_simh_hasbankedmemory
		cmp	data, #simh_stoptimer WZ
	if_z	jmp	#output_simh_stoptimer
		cmp	data, #simh_setbankselect WZ
	if_z	jmp	#output_simh_setbankselect
		mov	simh_status, #$01	' set error status?!
		jmp	#command_0

output_simh_params
		wrbyte	data, simh_params_ptr
		add	simh_params_ptr, #1
		sub	simh_params_cnt, #1 WZ
	if_nz	jmp	#command_0
		mov	simh_params_ptr, #0	' stop recording parameters
		' TODO: handle simh_command
		jmp	#command_0

output_simh_reset
		mov	simh_params_ptr, #0	' stop recording or replying parameters
		mov	simh_params_cnt, #0
		mov	simh_command, data
		mov	simh_status, #0
		jmp	#command_0

output_simh_setbankselect
		mov	simh_command, data
		mov	simh_params_ptr, #simh_params
		mov	simh_params_cnt, #1	' expect one more output: the bank number
		jmp	#command_0

output_simh_hasbankedmemory
		mov	simh_command, data
		mov	simh_status, #0
		jmp	#command_0

output_simh_stoptimer
		mov	simh_command, data
		mov	simh_status, #0
		jmp	#command_0

'*******************************************
'
' Handle hard disk command
'
ouput_hdsk
		cmp	hdsk_dir, #dir_out	WZ
	if_z	tjnz	hdsk_params_ptr, #output_hdsk_params

		cmp	data, #hdsk_reset	WZ
	if_z	jmp	#output_hdsk_reset
		cmp	data, #hdsk_read	WZ
	if_z	jmp	#output_hdsk_read
		cmp	data, #hdsk_write	WZ
	if_z	jmp	#output_hdsk_write
		cmp	data, #hdsk_param	WZ
	if_z	jmp	#output_hdsk_drive
		mov	hdsk_status, #$01		' set error status?!
		jmp	#command_0

'*******************************************
'
' Write port value to RAM at hdsk_params_ptr
'
output_hdsk_params
		wrbyte	data, hdsk_params_ptr
		add	hdsk_params_ptr, #1
		sub	hdsk_params_cnt, #1 WZ
	if_nz	jmp	#command_0
		mov	hdsk_params_ptr, #0		' stop recording parameters

		' handle hdsk_command
		cmp	hdsk_command, #hdsk_param	WZ
	if_z	jmp	#output_hdsk_params_drive
		cmp	hdsk_command, #hdsk_read	WZ
	if_z	jmp	#output_hdsk_read_dma
		cmp	hdsk_command, #hdsk_write	WZ
	if_z	jmp	#output_hdsk_write_dma
		jmp	#command_0

'*******************************************
'
' Reset the hdsk interface
'
output_hdsk_reset
		mov	hdsk_command, data
		mov	hdsk_params_ptr, #0		' stop recording parameters
		mov	hdsk_params_cnt, #0
		mov	hdsk_status, #0
		mov	hdsk_dir, #dir_none		' no direction
		jmp	#command_0

'*******************************************
'
' Read and write expect 6 additional
' writes to the port
'
output_hdsk_read
output_hdsk_write
		mov	hdsk_command, data
		mov	hdsk_params_ptr, hdsk_drive
		mov	hdsk_params_cnt, #6		' drive, sector, track lo, track hi, DMA lo, DMA hi
		mov	hdsk_status, #1			' non-zero status while expecting parameters?
		mov	hdsk_dir, #dir_out
		jmp	#command_0

'*******************************************
'
' Params expects one more out, the drive
' number, to be written to the port
'
output_hdsk_drive
		mov	hdsk_command, data
		mov	hdsk_params_ptr, hdsk_drive
		mov	hdsk_params_cnt, #1		' expect drive number
		mov	hdsk_dir, #dir_out
		jmp	#command_0

'*******************************************
'
' Params returns 19 bytes on the following
' reads from the port
output_hdsk_params_drive
		cmp	data, #8	WZ, WC		' we support drives 0 to 7
	if_ae	jmp	#:error
		mov	hdsk_params_ptr, hdsk_dpb
		mov	hdsk_params_cnt, #19		' 19 bytes to read
		mov	hdsk_status, #0
		mov	hdsk_dir, #dir_in
		jmp	#command_0
:error
		mov	hdsk_status, #$ff
		jmp	#command_0

'*******************************************
'
' Wait until the XMM driver clears the
' long[xmm_command]
'
xmm_wait_busy
:wait		rdlong	t1, xmm_command	WZ
	if_nz	jmp	#:wait
xmm_wait_busy_ret
		ret

'*******************************************
'
' Execute the hdsk read
'
output_hdsk_read_dma
		cmp	iobuff_dirty, #0 WZ		' flush a pending write
	if_nz	call	#sd_write_lba
		call	#sd_read_lba
		mov	t1, iobuff_ptr			' 512 byte buffer
		add	t1, sector_offs			' + 128 byte sector offset
		wrlong	t1, xmm_hubaddr			' set hub address of 128 byte sector
		rdbyte	t1, hdsk_dma_hi			' get DMA address high
		shl	t1, #8				' * 256
		rdbyte	t2, hdsk_dma_lo			' get DMA address low
		add	t1, t2				' + DMA address low
		wrlong	t1, xmm_ramaddr			' as XMM RAM address
		mov	t1, #128			' 128 bytes
		wrlong	t1, xmm_blocklen		' blocklen to write
		mov	t1, #"W"			' XMM write command
		wrlong	t1, xmm_command
		call	#xmm_wait_busy
		mov	hdsk_status, #0			' clear status
		jmp	#command_0

'*******************************************
'
' Execute the hdsk write
'
output_hdsk_write_dma
		call	#sd_read_lba			' make sure we have the sector
		mov	t1, iobuff_ptr			' 512 byte buffer
		add	t1, sector_offs			' + 128 byte sector offset
		wrlong	t1, xmm_hubaddr			' set hub address of 128 byte sector
		rdbyte	t1, hdsk_dma_hi			' get DMA address high
		shl	t1, #8				' * 256
		rdbyte	t2, hdsk_dma_lo			' get DMA address low
		add	t1, t2				' + DMA address low
		wrlong	t1, xmm_ramaddr			' as XMM RAM address
		mov	t1, #128			' 128 bytes
		wrlong	t1, xmm_blocklen		' blocklen to read
		mov	t1, #"R"			' XMM read command
		wrlong	t1, xmm_command
		call	#xmm_wait_busy
		mov	time_dirty, CNT
		add	time_dirty, clk_freq
		mov	iobuff_dirty, #1
		mov	hdsk_status, #0			' clear status
		jmp	#command_0

'*******************************************
'
' Wait while the SD driver is busy
'
sd_wait_busy
:wait		rdlong	t1, sd_command
		cmp	t1, t2			WZ
	if_z	jmp	#:wait
sd_wait_busy_ret
		ret

'*******************************************
'
' Read a logical block address from SD
'
sd_read_lba
		rdbyte	t1, hdsk_drive			' get drive number
		and	t1, #7				' we support 8 drives, no more
		shl	t1, #2				' * 4
		add	t1, hdsk_lba_drives		' offset of the per drive LBA0
		rdlong	lba0, t1
		rdbyte	t1, hdsk_track_hi		' t1 = track number high
		shl	t1, #8				' track number high * 256
		rdbyte	t2, hdsk_track_lo		' t2 = track low
		add	t1, t2				' + track number low
		shl	t1, #5				' * 32 (32 sectors per track)
		rdbyte	t2, hdsk_sector			' t3 = sector
		add	t1, t2				' + sector number
		mov	sector_offs, t1
		and	sector_offs, #3			' mask sector offset
		shl	sector_offs, #7			' * 128
		shr	t1, #2				' sector number / 4 (read 512 byte sectors)
		add	t1, lba0			' + first sector of this drive
		shl	t1, #8				' make sure bits 31..24 are zero
		shr	t1, #8
		mov	lba, t1
		cmp	lba, lba_iobuff		WZ	' == buffered lba?
	if_z	jmp	sd_read_lba_ret
		wrlong	lba, sd_blocknr
		mov	lba_iobuff, lba
		wrlong	iobuff_ptr, sd_address
		mov	t1, #"N"
		wrlong	t1, xmm_command
		call	#xmm_wait_busy
		mov	t1, #"r"			' spi read command
		mov	t2, t1
		wrlong	t1, sd_command			' please fetch this sector
		call	#sd_wait_busy
		mov	t1, #"F"
		wrlong	t1, xmm_command
		call	#xmm_wait_busy
sd_read_lba_ret
		ret

'*******************************************
'
' Write a logical block address to SD
'
sd_write_lba
		wrlong	lba, sd_blocknr
		mov	iobuff_dirty, #0		' buffer is clean
		wrlong	iobuff_ptr, sd_address
		mov	t1, #"N"
		wrlong	t1, xmm_command
		call	#xmm_wait_busy
		mov	t1, #"w"
		mov	t2, t1
		wrlong	t1, sd_command			' please store this sector
		call	#sd_wait_busy
		mov	t1, #"F"
		wrlong	t1, xmm_command
		call	#xmm_wait_busy
sd_write_lba_ret
		ret


'*******************************************
'
' Read i/o port to read from
'
input
		andn	iocmd, #$ff			' clear the iocmd data bits 7..0
		mov	port, iocmd			' extract port from bits 23..8
		shr	port, #8			' to bits 15..0
		and	port, #$ff			' mask lower bits 7..0 only

		cmp	port, #simh_port	WZ
	if_z	jmp	#input_simh
		cmp	port, #hdsk_port	WZ
	if_z	jmp	#input_hdsk
		cmp	port, #constat		WZ
	if_z	jmp	#input_constat
		cmp	port, #condata		WZ
	if_z	jmp	#input_condata
		cmp	port, #punstat		WZ
	if_z	jmp	#input_punstat
		cmp	port, #pundata		WZ
	if_z	jmp	#input_pundata
#ifdef HAVE_KBD
		cmp	port, #keystat		WZ
	if_z	jmp	#input_keystat
		cmp	port, #keydata		WZ
	if_z	jmp	#input_keydata
#endif
		jmp	#command_0

'*******************************************
'
' Read from the simh port
'
input_simh
		tjnz	simh_params_ptr, #input_simh_params
		or	iocmd, simh_status
		jmp	#command_0

'*******************************************
'
' Read port value from RAM at simh_params_ptr
'
input_simh_params
		rdbyte	data, simh_params_ptr
		add	simh_params_ptr, #1
		or	iocmd, data
		sub	simh_params_cnt, #1	WZ
	if_z	mov	simh_params_ptr, #0		' stop replying parameters
		jmp	#command_0

'*******************************************
'
' Read from the hdsk port
'
input_hdsk
		cmp	hdsk_dir, #dir_in	WZ
	if_z	tjnz	hdsk_params_ptr, #input_hdsk_params
		or	iocmd, hdsk_status
		jmp	#command_0

'*******************************************
'
' Read port value from RAM at hdsk_params_ptr
'
input_hdsk_params
		rdbyte	data, hdsk_params_ptr
		add	hdsk_params_ptr, #1
		or	iocmd, data
		sub	hdsk_params_cnt, #1	WZ
	if_z	mov	hdsk_params_ptr, #0		' stop replying parameters
		jmp	#command_0

'*******************************************
'
' Read the console status
'
input_constat
		rdlong	t1, rx_head
		rdlong	t2, rx_tail
		cmp	t1, t2			WZ
#ifdef HAVE_KBD
input_keystat
		rdlong	t1, kbd_head
		rdlong	t2, kbd_tail
	if_z	cmp	t1, t2			WZ
#endif HAVE_KBD
	if_z	mov	data, #%00000010	' no data available
	if_nz	mov	data, #%00000011	' data is available
input_constat_exit
		or	iocmd, data
		jmp	#command_0

'*******************************************
'
' Read the console data
'
' returns 0 if no data is available
'
input_condata
		mov	data, #0		' no data available
		rdlong	t1, rx_head
		rdlong	t2, rx_tail
		cmp	t1, t2		WZ
	if_z	jmp	#:kbd
		mov	addr, rx_buffer
		add	addr, t2
		rdbyte	data, addr
		add	t2, #1
		and	t2, #63
		wrlong	t2, rx_tail
		jmp	#input_constat_exit
:kbd
#ifdef HAVE_KBD
input_keydata
		rdlong	t1, kbd_head
		rdlong	t2, kbd_tail
		mov	t3, t2
		cmp	t1, t2		WZ
	if_z	jmp	#input_constat_exit
		shl	t2, #1
		add	t2, kbd_keys
		rdword	data, t2		' read just the lower byte
		add	t3, #1
		and	t3, #15
		wrlong	t3, kbd_tail
	
		' process key codes
		test	data, kbd_ctrl	WZ	' is it a control code?
	if_nz	and	data, #$1f		' make it $00..$1f
	if_nz	jmp	#input_constat_exit
		cmp	data, #200	WZ	' backspace?
	if_z	mov	data, #8		' backspace character
	if_z	jmp	#input_constat_exit
		cmp	data, #201	WZ	' is it delete?
	if_z	mov	data, #127		' Delete character
	if_z	jmp	#input_constat_exit
		cmp	data, #203	WZ	' upper code for Esc?
	if_z	mov	data, #27		' Escape character
	if_z	jmp	#input_constat_exit
		cmp	data, #128	WZ, WC	' don't return other codes
	if_ae	mov	data, #0
#endif HAVE_KBD
		jmp	#input_constat_exit

'*******************************************
'
' Read the paper punch status
'
input_punstat
		mov	data, #%00000011	' always allow to output to punch
		jmp	#input_constat_exit

'*******************************************
'
' Read the paper punch data
'
' returns 0 if no data is available
'
input_pundata
		mov	data, #0		' no data available
		jmp	#input_constat_exit

'*******************************************
'
' Break - issued when the Z80 encounters
' a HALT instruction.
'
break
		mov	t1, iocmd			' read the address of the register file
		shr	t1, #8				' bits 23..8 to 15..0
		and	t1, low_16bits			' mask only bits 15..0
		mov	strptr, msg_break_ptr	

		mov	t2, #8
:loop_r16	call	#conout_string
		rdword	hexval, t1			' read BC, DE, HL, AF, BC', DE', HL', AF'
		call	#conout_hex_w
		add	t1, #2
		djnz	t2, #:loop_r16

		call	#conout_string			' new line and print "R:"
		rdbyte	hexval, t1			' R
		add	t1, #1
		rdbyte	t2, t1				' R2
		and	hexval, #%01111111		' R register's bit #7
		or	hexval, t2			' does never change from the last written value
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdbyte	hexval, t1			' IFF1+2/IM
		mov	t2, t1				' save value
		and	hexval, #1			' IFF1 (interrupt flip-flop 1)
		call	#conout_hex_b

		call	#conout_string
		mov	hexval, t2			' IFF1+2/IM
		shr	hexval, #1			' shift bit #1 to #0
		and	hexval, #1			' IFF2 (interrupt flip-flop 2)
		call	#conout_hex_b

		call	#conout_string
		mov	hexval, t2			' IFF1+2/IM
		shr	hexval, #2			' shift bits #3..2 to #1..0
		and	hexval, #3			' IM (interrupt mode)
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdbyte	hexval, t1			' I (interrupt vector register)
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdword	hexval, t1			' IX
		call	#conout_hex_w
		add	t1, #2

		call	#conout_string
		rdword	hexval, t1			' IY
		call	#conout_hex_w
		add	t1, #2

		call	#conout_string
		rdword	hexval, t1			' SP
		call	#conout_hex_w
		add	t1, #2

		call	#conout_string
		rdword	hexval, t1			' PC
		mov	t3, hexval
		call	#conout_hex_w
		call	#conout_string
		add	t1, #2				' skip to opcode

		wrlong	t3, xmm_ramaddr			' as XMM ram address
		mov	t2, #4				' 4 bytes
		wrlong	t2, xmm_blocklen		' blocklen to read
		mov	t2, #"R"			' XMM read command
		wrlong	t2, xmm_command
:wait		rdlong	t2, xmm_command		WZ
	if_nz	jmp	#:wait

		rdbyte	hexval, t1
		call	#conout_hex_b

		call	#conout_crlf

		jmp	#command_0

		byte	0[$1f0-$]

		fit	$1f0

simh_params_	byte	0,0,0,0,0,0,0,0		' parameters of command

lba_drive_a	long	0			' first sector of drive A:
lba_drive_b	long	0			' first sector of drive B:
lba_drive_c	long	0			' first sector of drive C:
lba_drive_d	long	0			' first sector of drive D:
lba_drive_e	long	0			' first sector of drive E:
lba_drive_f	long	0			' first sector of drive F:
lba_drive_g	long	0			' first sector of drive G:
lba_drive_h	long	0			' first sector of drive H:

hdsk_params_	byte
hdsk_drive_	byte	0
hdsk_sector_	byte	0
hdsk_track_lo_	byte	0
hdsk_track_hi_	byte	0
hdsk_dma_lo_	byte	0
hdsk_dma_hi_	byte	0

'Hard Disk Parameter Blocks
'Each DPB is 17 bytes as defined for DPBs in CP/M followed by 2 bytes sector size as used by SIMH
hdsk_dpb_	byte			' 8MByte Default SIMH Altair HDSK params
:spt_low	byte	$20		' sectors per track (low byte)
:spt_high	byte	$00		' sectors per track (high byte)
:bsh		byte	$05		' data allocation Block SHift factor
:blm		byte	$1F		' data allocation block mask
:exm		byte	$01		' EXtent Mask
:dsm_low	byte	$F9		' maximum data block number (low_byte)
:dsm_high	byte	$07		' maximum data block number (high_byte)
:drm_low	byte	$FF		' total number of directory entries (low byte)
:drm_high	byte	$03		' total number of directory entries (high byte)
:al0		byte	$FF		' determine reserved directory blocks
:al1		byte	$00		' determine reserved directory blocks
:cks_low	byte	$00		' size of directory ChecK vector (low byte)
:cks_high	byte	$00		' size of directory ChecK vector (high byte)
:off_low	byte	$06		' number of reserved tracks (offset) (low byte)
:off_high	byte	$00		' number of reserved tracks (offset) (high byte)
:psh		byte	$00		' Physical record SHift factor, CP/M 3
:phm		byte	$00		' Physical record Mask, CP/M 3
:ss_low		byte	$80		' Sector Size (low byte)
:ss_high	byte	$00		' Sector Size (high byte)

msg_break	byte	"BC:",0
		byte	" DE:",0
		byte	" HL:",0
		byte	" AF:",0
		byte	" BC'",0
		byte	" DE'",0
		byte	" HL'",0
		byte	" AF'",0
		byte	13,10, "R:",0
		byte	" IFF1:",0
		byte	" IFF2:",0
		byte	" IM:",0
		byte	" I:",0
		byte	" IX:",0
		byte	" IY:",0
		byte	" SP:",0
		byte	" PC:",0
		byte	" = ",0

